# MIT License
# Copyright (c) 2023 sikino
# See LICENSE file in the project root for full license information.

# Detail see https://slpr.sakura.ne.jp/sikinote/docs/numeric/ivp/irk/

import numpy as np
import copy
from scipy.optimize import root


# 3-stage 6-order Gauss-Legendre Butcher tableau coefficients
c_GL = np.array([0.5 - 0.1 * np.sqrt(15.0), 0.5, 0.5 - 0.1 * np.sqrt(15.0)], dtype=np.float64)
a_GL = np.array(
    [
        [5.0 / 36.0, 2.0 / 9.0 - np.sqrt(15) / 15.0, 5.0 / 36.0 - np.sqrt(15.0) / 30.0],
        [5.0 / 36.0 + np.sqrt(15.0) / 24.0, 2.0 / 9.0, 5.0 / 36.0 - np.sqrt(15.0) / 24.0],
        [5.0 / 36.0 + np.sqrt(15.0) / 30.0, 2.0 / 9.0 + np.sqrt(15) / 15.0, 5.0 / 36.0],
    ],
    dtype=np.float64,
)
b_GL = np.array([5.0 / 18.0, 4.0 / 9.0, 5.0 / 18.0], dtype=np.float64)


# Right hand side of the differential equations
def ivp_func(x: float, y: np.ndarray, omega) -> np.ndarray:
    N = y.shape[0]
    res = np.zeros(N)
    res[0] = y[1]
    res[1] = -(omega * omega) * y[0]

    return res


# function g (NxP dimension) in https://slpr.sakura.ne.jp/sikinote/docs/numeric/ivp/irk/
def gfunc(
    k: np.ndarray,
    h: float,
    xin: float,
    yin: np.ndarray,
    omega: float,
) -> np.ndarray:
    N = yin.shape[0]
    P = c_GL.shape[0]
    NP = N * P

    g = np.zeros(NP)
    for i in range(P):
        x = xin + c_GL[i] * h
        ki = k[N * i : N * i + N]
        tmp = np.zeros(N)
        for j in range(P):
            kj = k[N * j : N * j + N]
            tmp = tmp + a_GL[i, j] * kj
        y = yin + h * tmp
        g[N * i : N * i + N] = ki - ivp_func(x, y, omega)

    return g


# Initial guess of k (NxP dimension) in https://slpr.sakura.ne.jp/sikinote/docs/numeric/ivp/irk/
def k_initial(
    xin: float,
    yin: np.ndarray,
    omega: float,
) -> np.ndarray:
    N = yin.shape[0]
    P = c_GL.shape[0]
    NP = N * P
    kin = np.zeros(NP)
    for i in range(P):
        x = xin
        y = yin
        kin[N * i : N * i + N] = ivp_func(x, y, omega)

    return kin


if __name__ == "__main__":
    N = 2  # Number of 1st order differencial equations
    P = c_GL.shape[0]  # Num of stages

    NP = N * P  # matrix dimension (NP x NP)

    omega = 1.0  # argment in the equation
    h = 0.01  # step size
    Nstep = 1000  # repetition of step x=[x0, x0+h*Nstep]

    # Initial conditions
    x0 = 0.0  # Initial value
    y0 = np.zeros(N)
    y0[0] = 0.0  # Initial value
    y0[1] = 1.0  # Initial value

    x = copy.copy(x0)
    y = copy.copy(y0)
    for j in range(Nstep):
        # -- IRK step --
        k = k_initial(x, y, omega)  # Initial guess for Newton-Raphson method
        args = ()
        #  Solve nonlinear simultaneous equations using scipy.optimize.root
        solution = root(gfunc, k, args=(h, x, y, omega), tol=6e-6, method="krylov")
        k = solution.x
        # ----
        tmp = np.zeros(N)
        for i in range(P):
            ki = k[N * i : N * i + N]
            tmp = tmp + b_GL[i] * ki

        # Update solutions
        y = y + h * tmp
        x = x + h

    print("Result: x, y[0], y[1] = ", x, y[0], y[1])
